import networkx as nx
from matplotlib import pyplot as plt
import numpy as np
from numba import jit


def plot_matrix(A, labels=False, all_ticks=False, extent=None, aspect=None):
    plt.imshow(A, cmap="autumn_r", extent=extent, aspect=aspect)

    if labels:
        for (j, i), label in np.ndenumerate(A):
            plt.text(i, j, label, ha='center', va='center')

    if all_ticks:
        plt.yticks(np.arange(len(A)))
        plt.xticks(np.arange(len(A)))


def info_graph(A):
    G = nx.from_numpy_array(A)
    print("Average number of edges : ", sum(sum(A)) / len(A))
    # print("Sigma Coefficient (sw if >1) : ", nx.sigma(G))
    # print("Omega Coefficient (sw if 0) : ", nx.omega(G))
    plt.figure()
    plot_matrix(A)
    plt.figure()
    plt.plot(nx.degree_histogram(nx.from_numpy_array(A)))
    plt.xlabel("Degree")
    plt.ylabel("Frequency")


# Implementation of copying an array to use numba
@jit
def copyto_numba(a, b):
    N = len(a)
    for i in range(N):
        b[i] = a[i]


@jit
def copyto_numba_array(A, B):
    N = len(A)
    for i in range(N):
        for j in range(N):
            B[i, j] = A[i, j]


# Swap two node's label in a network. It does not change the network but it's matrix representation
@jit
def swap(A_new, buf, i, j):
    copyto_numba(A_new[i, :], buf)
    # np.copyto(buf, A_new[i,:])
    A_new[i, :] = A_new[j, :]
    A_new[j, :] = buf

    copyto_numba(A_new[:, i], buf)
    # np.copyto(buf, A_new[:,i])
    A_new[:, i] = A_new[:, j]
    A_new[:, j] = buf
    return A_new


# Create a reordered matrix along new labels
def reorder(partitions, A):
    order_ = np.ndarray.flatten(partitions)
    A_ord_line = np.zeros_like(A)
    for i in range(len(order_)):
        A_ord_line[i] = A[order_[i]]
    A_ord_col = np.zeros_like(A)
    for i in range(len(order_)):
        A_ord_col[:, i] = A_ord_line[:, order_[i]]
    return A_ord_col


# Reorder a list of matrix along a list of labels
def reorder_list(order, l):
    return np.array([l[i] for i in order])


# Shuffle the label of a network. The network remains the same but not its incidency matrix.
def shuffle(A):
    A_new = np.copy(A)
    arr = np.arange(len(A_new))
    np.random.shuffle(arr)
    A_new = reorder(arr, A_new)
    return A_new



from numba import jit
from gmap.mapping import Hardware
from gmap.matrix_generator import create_communities
import numpy as np


@jit
def cost_update_mask_jit(A, Mask, i, j):
    if i == j: return 0
    A_i_, A_j_, A__i, A__j, A_ij, A_ji, A_ii, A_jj = \
        A[i, :] > 0, A[j, :] > 0, A[:, i] > 0, A[:, j] > 0, A[i, j] > 0, A[j, i] > 0, A[i, i] > 0, A[j, j] > 0
    update = (((A_j_ - A_i_) * Mask[i, :]).sum()
              + ((A_i_ - A_j_) * Mask[j, :]).sum()
              + ((A__j - A__i) * Mask[:, i]).sum()
              + ((A__i - A__j) * Mask[:, j]).sum()
              + ((A_ii + A_jj - 2 * A_ji) * Mask[i, i])
              + ((A_jj + A_ii - 2 * A_ij) * Mask[j, j])
              + ((A_ij + A_ji - 2 * A_jj) * Mask[i, j])
              + ((A_ji + A_ij - 2 * A_ii) * Mask[j, i]))

    return update


class Hardware_multicore(Hardware):
    def __init__(self, n_total, core):
        super(Hardware_multicore, self).__init__(n_total)  # important !
        self.Mask = 1 - 1 * create_communities(n_total, core)

    def update_cost(self, connectivity_matrix, a, b):
        return cost_update_mask_jit(connectivity_matrix, self.Mask, a, b)

    def cost(self, connectivity_matrix):
        return np.sum((connectivity_matrix > 0) * self.Mask)


class Hardware_generic(Hardware):
    def __init__(self, n_neurons_core, n_core, n_fanI=0, n_fanO=0):
        self.n_total = n_neurons_core * n_core
        self.n_neurons_core = n_neurons_core
        self.n_core = n_core
        self.n_fanI = n_fanI
        self.n_fanO = n_fanO
        super(Hardware_generic, self).__init__(self.n_total)  # important !
        self.Mask = 1 - 1 * create_communities(self.n_total, self.n_core)

    def violated_fan(self, connectivity_matrix, axis, max_fan):
        ext_fan = self.Mask * (connectivity_matrix > 0)
        sum_fan = np.sum(ext_fan, axis=axis)
        return np.sum( (sum_fan-max_fan) * (sum_fan > max_fan) )

    def cost(self, connectivity_matrix):
        violated_FI = self.violated_fan(connectivity_matrix, 0, self.n_fanI)  # Sum along axis 0 (vertically = fan-in)
        violated_FO = self.violated_fan(connectivity_matrix, 1,
                                        self.n_fanO)  # Sum along axis 1 (horizontally = fan-out)
        return violated_FI + violated_FO


class Hardware_DYNAPS(Hardware):
    def __init__(self, N, F, C, K, M=None, alpha=None):
        # Either one describe DYNAPS with M or alpha.
        assert M is None or alpha is None, "Error : Both M and alpha are not None"
        assert M is not None or alpha is not None, "Error : Both M and alpha are None"
        if alpha is not None:
            M = int(np.sqrt((F * np.log2(alpha * N)) / (alpha * np.log2(alpha * C))))

        # Check conditions
        assert F > M, "Error : Condition F > M not respected"
        assert F / M <= N / C, "Error : Condition F/M <= N/C not respected"
        assert M <= C, "Error : Condition M <= C not respected"

        self.N = N
        self.F = F
        self.M = M
        self.C = C
        self.K = K
        super(Hardware_DYNAPS, self).__init__(self.N)  # important !

        self.mat_FI = [[1.0 if j == i // self.C else 0.0 for j in range(self.N // self.C)] for i in range(self.N)]

    def violated_mem_sender(self, connectivity_matrix):
        # Compute the fan in of each neuron to each cluster
        n_FI = np.dot(connectivity_matrix>0, self.mat_FI)

        # Each neuron can communicate with at most M neurons within a cluster
        violated_M = np.sum((n_FI - self.M) * (n_FI > self.M))

        # Each neuron can communicate with at most F/M cluster
        n_FI_cluster = np.count_nonzero(n_FI, axis=1) # compute the number of cluster a neuron is communicating with.
        max_FI_cluster = self.F // self.M
        violated_F = np.sum((n_FI_cluster - max_FI_cluster) * (n_FI_cluster > max_FI_cluster))

        return violated_M + violated_F

    def violated_mem_receiver(self, connectivity_matrix):
        ''' Each cluster can have at most K combination of fan-in, the K tags '''
        violated_K = 0
        for i in range(self.N//self.C) :
            # compute the number of different combination of neurons, the number of required tags.
            A = (connectivity_matrix>0)[i*self.C:(i+1)*self.C]
            nonzero_rows = A[np.any(A != 0, axis=1)]
            unique_nonzero_rows = np.unique(nonzero_rows, axis=0)
            max_K = unique_nonzero_rows.shape[0]

            # Compare it with the actual max number of tags.
            violated_K += (max_K-self.K)*(max_K > self.K)

        return violated_K


    def cost(self, connectivity_matrix):
        return self.violated_mem_sender(connectivity_matrix) + self.violated_mem_receiver(connectivity_matrix)



import random
from simanneal import Annealer
from gmap.utils import *


class Hardware_Annealer(Annealer):
    def __init__(self, state, update_move, cost, debug):
        self.update_move = update_move
        self.cost = cost
        super(Hardware_Annealer, self).__init__(state)  # important!
        if not debug: self.update = lambda *args, **kwargs: None

    def move(self):
        a = random.randint(0, len(self.state) - 1)
        b = random.randint(0, len(self.state) - 1)
        self.state, dE = self.update_move(self.state, a, b)
        self.user_exit = self.best_energy is not None and self.best_energy <= 0
        return dE

    def energy(self):
        return self.cost(self.state)


class Hardware:
    def __init__(self, n_total):
        self.n_total = n_total
        self.buf = np.empty(n_total)  # pre-allocated buffer to avoid re-alloc at each swap.
        pass

    def cost(self, connectivity_matrix):
        pass

    def update_cost(self, connectivity_matrix, a, b):
        return None

    def update_move(self, connectivity_matrix, a, b):
        dE = self.update_cost(connectivity_matrix, a, b)
        if dE is None:
            E_ini = self.cost(connectivity_matrix)
            connectivity_matrix = swap(connectivity_matrix, self.buf, a, b)
            dE = self.cost(connectivity_matrix) - E_ini
        else:
            connectivity_matrix = swap(connectivity_matrix, self.buf, a, b)
        return connectivity_matrix, dE

    def mapping(self, connectivity_matrix, minutes=0.5, debug=False):
        # pre-check
        if len(connectivity_matrix) > self.n_total:
            return connectivity_matrix, float('inf')

        # pre-treatment
        pad = self.n_total - len(connectivity_matrix)
        connectivity_matrix = np.pad(connectivity_matrix, [(0, pad), (0, pad)], mode='constant')

        # solving the mapping
        sa = Hardware_Annealer(connectivity_matrix, self.update_move, self.cost, debug=debug)

        # Searching for the good params
        if debug :print("Searching for the good optimization params...")
        param = sa.auto(minutes=minutes)
        param['updates'] = 10
        sa.set_schedule(param)

        # Solving the mapping
        if debug : print("Solving the mapping...")
        mapping, violated_constrains = sa.anneal()

        if debug :
            if violated_constrains == 0: print("Mappable !")
            else: print("Not Mappable, violated : ", violated_constrains)

        return mapping, violated_constrains



import networkx as nx
import numpy as np
from scipy.stats import norm


# Function that create a random graph
def create_random(N, k_avg):
    ret = np.zeros((N, N))
    A = np.random.uniform(size=(N, N))
    ret[:N, :N] = np.array(1 * (A < (k_avg / N)))
    return ret


# Create a Mask representing a network of N nodes with C communities
def create_communities(N, C):
    A = np.zeros((N, N))
    for i in range(C):
        A[(N // C) * i:(N // C) * (i + 1), (N // C) * i:(N // C) * (i + 1)] = 1
    return A.astype(bool)


# Function that create a Small wolrd Watts Stoggart network.
def create_WS(N, k_avg, p_drop=0.3):
    A = np.zeros((N, N))
    G = nx.connected_watts_strogatz_graph(N, k_avg, p_drop)
    A[:N, :N] = nx.to_numpy_array(G)  # [K:,K:]
    return A


# Create a Barabasi-Albert network
def BA(N, M0, m):
    """
        N, number of nodes in the final network.
        M0, initial connected network of M0 nodes.
        m, Each new node is connected to m existing nodes.
    """
    assert (M0 < N)
    assert (m <= M0)

    # adjacency matrix
    AM = np.zeros((N, N))

    for i in range(0, M0):
        for j in range(i + 1, M0):
            AM[i, j] = 1
            AM[j, i] = 1

    # add 'c' node
    for c in range(M0, N):
        Allk = np.sum(AM)  # all  degree    Eki
        ki = np.sum(AM, axis=1)  # ki each degree for node i

        pi = np.zeros(c, dtype=np.float)  # probability
        for i in range(0, c):
            pi[i] = ki[i] / (Allk * 1.0)
        # print pi

        # connect m edges.
        for d in range(0, m):
            rand01 = np.random.random()  # [0,1.0)

            sumpi = 0.0
            for g in range(0, c):
                sumpi += pi[g]
                if sumpi > rand01:  # connect 'c' node with 'g' node.
                    if AM[c, g] == 0 and AM[g, c] == 0:
                        AM[c, g] += 1
                        AM[g, c] += 1
                        break

    return AM


# Build a Barabasi-Albert Network with same M0 = m and with the right average number of edges per node.
def create_BA(N, k_avg):
    M = int((-1 + 2 * N - ((1 - 2 * N) ** 2 - 4 * N * k_avg) ** (1 / 2)) // 2) + 1
    return BA(N, M, M)


def gaussian_connect(N, sigma, mu, tol=1e-6, quantile_max=30):
    alpha = 3 / quantile_max ** 2 - norm.cdf(-quantile_max) / norm.pdf(-quantile_max) * 3 / quantile_max
    assert mu ** 2 / 12 * (1 + alpha) < sigma ** 2, "Error: mu ** 2 / 12 > sigma ** 2"
    assert sigma ** 2 < N ** 2 / 12, "Error: sigma ** 2 > N ** 2 / 12"
    mu_target = mu
    sigma_target = sigma
    margin = sigma * np.sqrt(np.maximum(np.log(mu ** 2 / 2 / np.pi / sigma ** 2), 0))
    while True:
        cdf_diff = norm.cdf(- margin / sigma) - norm.cdf(- N / 2 / sigma)
        if margin:
            mu = np.sqrt(2 * np.pi) * sigma * np.exp(margin ** 2 / 2 / sigma ** 2)
        else:
            mu = mu_target / 2 / cdf_diff
        b = - mu * np.sqrt(2 / np.pi) * N / 2 * np.exp(- (N / 2) ** 2 / 2 / sigma ** 2) / mu_target
        c = 2 / 3 * margin ** 3 / mu_target - sigma_target ** 2
        if np.abs(2 * mu * cdf_diff + 2 * margin - mu_target) < tol and \
                np.abs(sigma ** 2 + b * sigma + c) < tol:
            break
        sigma = np.sqrt(- (c + b * sigma))
        if margin:
            margin = np.maximum(mu_target / 2 - mu * cdf_diff, 0)
        else:
            margin = sigma * np.sqrt(np.maximum(np.log(mu ** 2 / 2 / np.pi / sigma ** 2), 0))
    x, y = np.meshgrid(range(N), range(N))
    d = np.minimum(np.mod(x - y, N), np.mod(y - x, N))
    return np.random.rand(N, N) <= mu / np.sqrt(2 * np.pi) / sigma * np.exp(- d ** 2 / 2 / sigma ** 2)


# Create a graph in which the probability of having an edge between any given pair of nodes is a Guassian distribution with respect to the distance between them.
def create_gaussian_connect(N, k_avg):
    A = np.zeros((N, N))
    quantile_max = 30
    alpha = 3 / quantile_max ** 2 - norm.cdf(-quantile_max) / norm.pdf(-quantile_max) * 3 / quantile_max
    min_sigma = (k_avg ** 2 / 12 * (1 + alpha)) ** (1 / 2)
    max_sigma = (N ** 2 / 12) ** (1 / 2)
    g = gaussian_connect(N, (min_sigma + max_sigma) * (1 / 2), k_avg)
    A[:N, :N] = g
    return A


# Create a network with the number of connection depending of the function f along the distance
# example of use :
# A = create_distance_dependant(N, minus_x_square(N, 3) , circular = True)
def create_distance_dependant(N, func, circular=True):
    x, y = np.meshgrid(range(N), range(N))
    if circular:
        A = np.minimum(np.mod(x - y, N), np.mod(y - x, N))
    else:
        A = np.abs(x - y)
    return np.vectorize(func)(A)


def linear(x): return x


def uni_proba(x): return np.random.uniform(0, 1) < x


def minus_x_exponent(N, exp):
    return lambda x: uni_proba(((N - x) / N) ** exp)



from setuptools import setup, find_packages

setup(
    name='gmap-compiler',
    version='0.3',
    packages=find_packages(),
    install_requires=[
        'numpy',
        'matplotlib',
        'networkx',
        'numba',
        'scipy',
        'simanneal',
    ],
    author='Jimmy Weber',
    author_email='jimmy.weber@ini.ethz.ch',
    description='A versatile, easy-to-use and open-source compiler that can efficiently map any arbitrary connectivity matrix to various hardware architectures.',
    url='https://github.com/EIS-Hub/GMap',
)

# To update the package, run "python setup.py sdist"